# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

print "No automatic tests defined for Net::IRC. Maybe I ought to make some,\n",
      "but I'm a lot more interested in actually getting it working first.\n";

exit 0;
